@@ -21,6 +21,7 @@ from scalecast.Forecaster import (
     _colors_,
     ForecastError
 )
+import copy
 
 # LOGGING
 logging.basicConfig(filename="warnings.log", level=logging.WARNING)
@@ -101,24 +102,24 @@ class MVForecaster:
         self.freq = fs[0].freq
         self.n_series = len(fs)
         for i, f in enumerate(fs):
-            setattr(self, f'series{i+1}', {'y':f.y.copy(),
+            setattr(self, f'series{i+1}', {'y':f.y.copy().reset_index(drop=True),
                 'levely':f.levely.copy(),
                 'integration':f.integration})
             if i == 0:
-                self.current_dates = f.current_dates.copy()
+                self.current_dates = f.current_dates.copy().reset_index(drop=True)
             if merge_Xvars in ('union','u'):
                 if i == 0:
-                    self.current_xreg = {k:v.copy() for k, v in f.current_xreg.items() if not k.startswith('AR')}
+                    self.current_xreg = {k:v.copy().reset_index(drop=True) for k, v in f.current_xreg.items() if not k.startswith('AR')}
                     self.future_xreg = {k:v[:] for k, v in f.future_xreg.items() if not k.startswith('AR')}
                 else:
                     for k, v in f.current_xreg.items():
                         if not k.startswith('AR'):
-                            self.current_xreg[k] = v.copy()
+                            self.current_xreg[k] = v.copy().reset_index(drop=True)
                             self.future_xreg[k] = f.future_xreg[k][:]
             elif merge_Xvars in ('intersection','i'):
                 if i == 0:
-                    self.current_xreg = {k:v in f.current_xreg.items()}
-                    self.future_xreg = {k:v in f.future_xreg.items()}
+                    self.current_xreg = {k:v.copy().reset_index(drop=True) in f.current_xreg.items()}
+                    self.future_xreg = {k:v[:] in f.future_xreg.items()}
                 else:
                     for k, v in f.current_xreg.items():
                         if k not in self.current_xreg.keys():
@@ -141,9 +142,9 @@ class MVForecaster:
 
         if names is not None: # checking for these objects is how we know whether user supplied names later
             names = list(names)
-            globals()['name_series_map'] = {names[i]:[f'series{i+1}',f'y{i+1}'] for i in range(self.n_series)}
-            globals()['y_name_map'] = {f'y{i+1}':names[i] for i in range(self.n_series)}
-            globals()['series_name_map'] = {f'series{i+1}':names[i] for i in range(self.n_series)}
+            self.name_series_map = {names[i]:[f'series{i+1}',f'y{i+1}'] for i in range(self.n_series)}
+            self.y_name_map = {f'y{i+1}':names[i] for i in range(self.n_series)}
+            self.series_name_map = {f'series{i+1}':names[i] for i in range(self.n_series)}
 
     def __repr__(self):
         return """MVForecaster(
@@ -169,8 +170,8 @@ class MVForecaster:
             len(self.current_dates),
             self.n_series,
             [f'series{i+1}' for i in range(self.n_series)] 
-            if 'name_series_map' not in globals() 
-            else list(name_series_map.keys()),
+            if not hasattr(self,'name_series_map') 
+            else list(self.name_series_map.keys()),
             len(self.future_dates),
             list(self.current_xreg.keys()),
             self.test_length,
@@ -182,6 +183,27 @@ class MVForecaster:
             self.estimator,
             self.optimize_on)
 
+    def __copy__(self):
+        obj = type(self).__new__(self.__class__)
+        obj.__dict__.update(self.__dict__)
+        return obj
+
+    def __deepcopy__(self):
+        obj = type(self).__new__(self.__class__)
+        obj.__dict__.update(self.__dict__)
+        obj.__dict__ = copy.deepcopy(obj.__dict__)
+        return obj
+
+    def copy(self):
+        """ creates an object copy.
+        """
+        return self.__copy__() 
+
+    def deepcopy(self):
+        """ creates an object deepcopy.
+        """
+        return self.__deepcopy__()
+
     def add_sklearn_estimator(self,imported_module,called):
         """ adds a new estimator from scikit-learn not built-in to the forecaster object that can be called using set_estimator().
         be careful to choose regression models only.
@@ -391,10 +413,17 @@ class MVForecaster:
                 raise ForecastError.NoGrid(f'to tune, a grid must be loaded. tried to load a grid called {self.estimator}, but either the MVGrids.py file could not be found in the current directory, there is no grid with that name, or the dictionary values are not list-like. try ingest_grid() with a dictionary grid passed manually.')
 
         metrics = {f'y{i+1}_metric':[] for i in range(self.n_series)}
-        for i, v in self.grid.iterrows():
+        iters = self.grid.shape[0]
+        for i in range(iters):
             try:
                 # returns a dict
-                val_preds, val_ac = self._forecast(fcster=self.estimator,tune=True,dynamic_testing=dynamic_tuning,**v)
+                hp = {k:v[i] for k,v in self.grid.to_dict(orient='list').items()}
+                val_preds, val_ac = self._forecast(
+                    fcster=self.estimator,
+                    tune=True,
+                    dynamic_testing=dynamic_tuning,
+                    **hp
+                )
                 for series, a in val_ac.items():
                     metrics[series + '_metric'].append(globals()[self.validation_metric](a,val_preds[series]))
             except TypeError:
@@ -402,7 +431,7 @@ class MVForecaster:
             except Exception as e:
                 self.grid.drop(i, axis=0, inplace=True)
                 logging.warning(
-                    f"could not evaluate the paramaters: {dict(v)}. error: {e}"
+                    f"could not evaluate the paramaters: {hp}. error: {e}"
                 )
         metrics = pd.DataFrame(metrics)
         if metrics.shape[0] > 0:
@@ -556,8 +585,8 @@ class MVForecaster:
         if how == 'mean':
             self.optimize_on = 'mean'
             return
-        if 'name_series_map' in globals():
-            how = name_series_map[how][0]
+        if hasattr(self,'name_series_map'):
+            how = self.name_series_map[how][0]
         descriptive_assert(how.startswith('series') or how.startswith('y') or how == 'mean',ValueError,f'value passed to how not usable: {how}')
         self.optimize_on = how if how.startswith('series') else 'series{}'.format(how.split('y')[-1])
 
@@ -590,9 +619,6 @@ class MVForecaster:
                 if dict, the key must be the user-selected series name, 'series{n}' or 'y{n}' and key is list or int.
             **kwargs: treated as model hyperparameters and passed to _sklearn_imports_[model]()
 
-        Returns:
-            (float or list): The evaluated metric value on the validation set if tuning a model otherwise, the list of predictions.
-        
         >>> mvf.set_estimator('gbt')
         >>> mvf.manual_forecast(lags=3) # adds three lags for each series
         >>> mvf.manual_forecast(lags=[1,3]) # first and third lags added for each series
@@ -914,22 +940,22 @@ class MVForecaster:
         """
         def parse_str_arg(arg):
             if arg.startswith('series'):
-                series = [f'y{i}'.format(arg.split('series')[-1])]
-                labels = arg.copy()
+                series = ['y{}'.format(arg.split('series')[-1])]
+                labels = series.copy()
             elif arg.startswith('y'):
                 series = [arg]
                 labels = series.copy()
             else:
                 labels = [arg]
-                series = [name_series_map[arg][1]]
+                series = [self.name_series_map[arg][1]]
             return series, labels
 
         if series == 'all':
             series = [f'y{i+1}' for i in range(self.n_series)]
-            if 'name_series_map' in globals():
-                labels = list(name_series_map.keys())
+            if hasattr(self,'name_series_map'):
+                labels = list(self.name_series_map.keys())
             else:
-                lables = series.copy()
+                labels = series.copy()
         elif isinstance(series,str):
             series, labels = parse_str_arg(series)
         else:
@@ -991,7 +1017,7 @@ class MVForecaster:
         k = 0
         for i, s in enumerate(series):
             sns.lineplot(x=self.current_dates.to_list(),
-                y=getattr(self,'series{}'.format(s.split('y')[-1]))['y'].to_list() if not level else getattr(self,f'series{i+1}')['levely'][-len(self.current_dates):],
+                y = list(getattr(self,'series{}'.format(s.split('y')[-1]))['y' if not level else 'levely'])[-len(self.current_dates):],
                 label = f'{labels[i]} actual',
                 ax=ax,
                 color = _series_colors_[i])
@@ -1066,7 +1092,7 @@ class MVForecaster:
 
         k = 0
         for i, s in enumerate(series):
-            y = getattr(self,'series{}'.format(s.split('y')[-1]))['y'].to_list() if not level else getattr(self,s)['levely'][-len(self.current_dates):]
+            y = list(getattr(self,'series{}'.format(s.split('y')[-1]))['y' if not level else 'levely'])[-len(self.current_dates):]
             sns.lineplot(x=self.current_dates.to_list()[-include_train:],
                 y=y[-include_train:],
                 label = f'{labels[i]} actual',
@@ -1212,8 +1238,8 @@ class MVForecaster:
                     elif c == "OptimizedOn" and hasattr(self,'best_model'): 
                         if self.optimize_on == 'mean':
                             model_summary_sm['OptimizedOn'] = ['mean']
-                        elif 'series_name_map' in globals():
-                            model_summary_sm['OptimizedOn'] = [series_name_map[self.optimize_on]]
+                        elif hasattr(self,'series_name_map'):
+                            model_summary_sm['OptimizedOn'] = [self.series_name_map[self.optimize_on]]
                         else:
                             model_summary_sm['OptimizedOn'] = [self.optimize_on]
                         if hasattr(self,'optimize_metric'):
@@ -1362,4 +1388,92 @@ class MVForecaster:
             (DataFrame): the validation grid
         """
         return self.history[model]['grid_evaluated'].copy()
-        
\ No newline at end of file
+
+    def backtest(self,model,fcst_length='auto',n_iter=10):
+        """ runs a backtest of a selected evaluated model over a certain 
+        amount of iterations to test the average error if that model were 
+        implemented over the last so-many actual forecast intervals.
+        all scoring is dynamic to give a true out-of-sample result.
+        all metrics are specific to level data.
+        two results are extracted: a dataframe of actuals and predictions across each iteration and
+        a dataframe of test-set metrics across each iteration with a mean total as the last column.
+        these results are stored in the Forecaster object's history and can be extracted by calling 
+        `export_backtest_metrics()` and `export_backtest_values()`.
+
+        Args:
+            model (str): the model to run the backtest for. use the model nickname.
+            fcst_length (int or str): default 'auto'. 
+                if 'auto', uses the same forecast length as saved in the object currently.
+                if int, uses that as the forecast length.
+            n_iter (int): default 10. the number of iterations to backtest.
+                models will iteratively train on all data before the fcst_length worth of values.
+                each iteration takes one observation off the end to redo the cast until all of n_iter is exhausted.
+
+        Returns:
+            None
+        """
+        if fcst_length == 'auto':
+            fcst_length = len(self.future_dates)
+        fcst_length = int(fcst_length)
+        series, labels = self._parse_series('all')
+        mets = ['RMSE','MAE','R2','MAPE']
+        tuples = []
+        for s in labels:
+            for m in mets:
+                tuples.append((s,m)) # no list comprehension to preserve order and because the lengths of labels and mets are different
+        index = pd.MultiIndex.from_tuples(tuples,names=['series','metric'])
+        metric_results = pd.DataFrame(columns=[f'iter{i}' for i in range(1,n_iter+1)],
+            index=index)
+        value_results = pd.DataFrame()
+        for i in range(n_iter):
+            f = self.__deepcopy__()
+            if i > 0:
+                f.current_xreg = {k:pd.Series(v.values[:-i]) for k, v in f.current_xreg.items()}
+                f.current_dates = pd.Series(f.current_dates.values[:-i])
+                for k in range(f.n_series):
+                    getattr(f,f'series{k+1}')['y'] = pd.Series(getattr(f,f'series{k+1}')['y'].values[:-i])
+                    getattr(f,f'series{k+1}')['levely'] = getattr(f,f'series{k+1}')['levely'][:-i]
+                f.future_dates = pd.Series(f.future_dates.values[:-i])
+
+            f.set_test_length(fcst_length)
+            f.set_estimator(f.history[model]['Estimator'])
+            params = f.history[model]['HyperParams'].copy()
+            params['lags'] = f.history[model]['Lags']
+            params['normalizer'] = f.history[model]['Scaler']
+            f.history = {}
+            f.manual_forecast(**params)
+            test_mets = f.export_model_summaries()
+            test_preds = f.export_level_test_set_preds()
+            for s in labels:
+                for m in mets:
+                    metric_results.loc[(s,m),f'iter{i+1}'] = test_mets.loc[test_mets['Series'] == s,f'LevelTestSet{m}'].values[0]
+                value_results[f'{s}_iter{i+1}actuals'] = test_preds[f'{s}_actuals']
+                value_results[f'{s}_iter{i+1}preds'] = test_preds[f'{s}_{model}_test_preds']
+        metric_results['mean'] = metric_results.mean(axis=1)
+        self.history[model]['BacktestMetrics'] = metric_results
+        self.history[model]['BacktestValues'] = value_results
+        # TODO: value_results can be a dataframe subclass with a custom plot() method
+    
+    def export_backtest_metrics(self,model):
+        """ extracts the backtest metrics for a given model.
+        only works if `backtest()` has been called.
+
+        Args:
+            model (str): the model nickname to extract metrics for.
+
+        Returns:
+            (DataFrame): a copy of the backtest metrics.
+        """
+        return self.history[model]['BacktestMetrics'].copy()
+
+    def export_backtest_values(self,model):
+        """ extracts the backtest values for a given model.
+        only works if `backtest()` has been called.
+
+        Args:
+            model (str): the model nickname to extract values for.
+
+        Returns:
+            (DataFrame): a copy of the backtest values.
+        """
+        return self.history[model]['BacktestValues'].copy()
\ No newline at end of file
